import logging
import traceback
from datetime import datetime
from pathlib import Path

import click

from tulona.cli import params as p
from tulona.config.profile import Profile
from tulona.config.project import Project
from tulona.config.runtime import RunConfig
from tulona.exceptions import TulonaMissingPropertyError
from tulona.task.compare import CompareColumnTask, CompareRowTask, CompareTask
from tulona.task.ping import PingTask
from tulona.task.profile import ProfileTask
from tulona.task.scan import ScanTask
from tulona.util.filesystem import get_final_outdir

log = logging.getLogger()
log_formatter = logging.Formatter(
    "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
)

consloe_handler = logging.StreamHandler()
consloe_handler.setFormatter(log_formatter)
consloe_handler.setLevel(logging.INFO)
log.addHandler(consloe_handler)

log_dir = Path(Path().absolute(), "log")
log_dir.mkdir(parents=True, exist_ok=True)
log_file_fqn = Path(log_dir, f"tulona_{datetime.now().strftime('%Y%m%d%H%M%S')}.log")
file_handler = logging.FileHandler(log_file_fqn)
file_handler.setFormatter(log_formatter)
file_handler.setLevel(logging.DEBUG)
log.addHandler(file_handler)


# command: tulona
@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    epilog="Execute: tulona <command> -h/--help for more help with specific commands",
)
@click.pass_context
def cli(ctx):
    """Tulona compares data sources to find out differences"""


# command: tulona ping
@cli.command("ping")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
def ping(ctx, **kwargs):
    """Test connectivity to datasources"""

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)
        logging.getLogger("snowflake").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]

    ping_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare",
            "datasources": kwargs["datasources"].split(","),
        }
        ping_tasks.append(task_config)
    else:
        ping_tasks = [t for t in ctx.obj["project"]["task_config"] if t["task"] == "ping"]

    if len(ping_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number of ping tasks to execute: {len(ping_tasks)}")

    for tconf in ping_tasks:
        log.debug(f"Executing ping with task profile: {tconf}")
        PingTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            datasources=tconf["datasources"],
        ).execute()


# command: tulona scan
@cli.command("scan")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.compare
@p.sample_count
@p.composite
def scan(ctx, **kwargs):
    """Scan data sources to collect metadata"""

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    scan_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare",
            "datasources": kwargs["datasources"].split(","),
        }
        if kwargs["compare"]:
            task_config["compare"] = kwargs["compare"]
        if kwargs["sample_count"]:
            task_config["sample_count"] = kwargs["sample_count"]
        if kwargs["composite"]:
            task_config["composite"] = kwargs["composite"]
        scan_tasks.append(task_config)
    else:
        scan_tasks = [t for t in ctx.obj["project"]["task_config"] if t["task"] == "scan"]

    if len(scan_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number of scan tasks to execute: {len(scan_tasks)}")

    for tconf in scan_tasks:
        log.debug(f"Executing scan with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )

        ScanTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            final_outdir=final_outdir,
            compare=tconf["compare"] if "compare" in tconf else None,
            sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
            composite=tconf["composite"] if "composite" in tconf else None,
        ).execute()


# command: tulona profile
@cli.command("profile")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.compare
def profile(ctx, **kwargs):
    """Profile data sources to collect metadata [row count, column min/max/mean etc.]"""

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    profile_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare",
            "datasources": kwargs["datasources"].split(","),
        }
        if kwargs["compare"]:
            task_config["compare"] = kwargs["compare"]
        profile_tasks.append(task_config)
    else:
        profile_tasks = [
            t for t in ctx.obj["project"]["task_config"] if t["task"] == "profile"
        ]

    if len(profile_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number of profile tasks to execute: {len(profile_tasks)}")

    for tconf in profile_tasks:
        log.debug(f"Executing profile with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "profile_metadata.xlsx")

        ProfileTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            outfile_fqn=outfile_fqn,
            compare=tconf["compare"] if "compare" in tconf else None,
        ).execute()


# command: tulona compare-row
@cli.command("compare-row")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.sample_count
def compare_row(ctx, **kwargs):
    """Compares rows from two data entities"""

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    compare_row_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare",
            "datasources": kwargs["datasources"].split(","),
        }
        if kwargs["sample_count"]:
            task_config["sample_count"] = kwargs["sample_count"]
        compare_row_tasks.append(task_config)
    else:
        compare_row_tasks = [
            t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare-row"
        ]

    if len(compare_row_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number compare-row tasks to execute: {len(compare_row_tasks)}")

    for tconf in compare_row_tasks:
        log.debug(f"Executing compare-row with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "row_comparison.xlsx")

        CompareRowTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            outfile_fqn=outfile_fqn,
            sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
        ).execute()


# command: tulona compare-column
@cli.command("compare-column")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.composite
def compare_column(ctx, **kwargs):
    """
    Column name must be specified for task: compare-column
    by specifying 'compare_column' property in
    all the datasource[project] configs
    (check sample tulona-project.yml file for example)
    """

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    compare_column_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare-column",
            "datasources": kwargs["datasources"].split(","),
        }
        if kwargs["composite"]:
            task_config["composite"] = kwargs["composite"]
        compare_column_tasks.append(task_config)
    else:
        compare_column_tasks = [
            t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare-column"
        ]

    if len(compare_column_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number compare-column tasks to execute: {len(compare_column_tasks)}")

    for tconf in compare_column_tasks:
        log.debug(f"Executing compare-column with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "column_comparison.xlsx")

        CompareColumnTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            outfile_fqn=outfile_fqn,
            composite=tconf["composite"] if "composite" in tconf else None,
        ).execute()


# command: tulona compare
@cli.command("compare")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
@p.datasources
@p.sample_count
@p.composite
def compare(ctx, **kwargs):
    """
    Compare everything(profiles, rows and columns) for the given datasoures
    """

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    compare_tasks = []
    if kwargs["datasources"]:
        task_config = {
            "task": "compare",
            "datasources": kwargs["datasources"].split(","),
        }
        if kwargs["sample_count"]:
            task_config["sample_count"] = kwargs["sample_count"]
        if kwargs["composite"]:
            task_config["composite"] = kwargs["composite"]
        compare_tasks.append(task_config)
    else:
        compare_tasks = [
            t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare"
        ]

    if len(compare_tasks) == 0:
        raise RuntimeError(
            "Nothing to execute. Either define tasks in task_config section of the"
            " project config file or pass --datasources argument with values."
        )
    log.debug(f"Number compare tasks to execute: {len(compare_tasks)}")

    for tconf in compare_tasks:
        log.debug(f"Executing compare with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "comparison.xlsx")

        CompareTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            outfile_fqn=outfile_fqn,
            sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
            composite=tconf["composite"] if "composite" in tconf else None,
        ).execute()


# command: tulona run
@cli.command("run")
@click.pass_context
# @p.exec_engine
@p.outdir
@p.verbose
def run(ctx, **kwargs):
    """Run all tasks defined by `task_config` attribute in the project config file"""

    if kwargs["verbose"]:
        logging.getLogger("tulona").setLevel(logging.DEBUG)

    prof = Profile()
    proj = Project()

    ctx.obj = ctx.obj or {}
    ctx.obj["project"] = proj.load_project_config()
    ctx.obj["profile"] = prof.load_profile_config()[ctx.obj["project"]["name"]]
    ctx.obj["runtime"] = RunConfig(options=kwargs, project=ctx.obj["project"])

    if "task_config" not in ctx.obj["project"]:
        raise TulonaMissingPropertyError(
            "Attribute `task_config` is not defined in project config"
        )

    # Override config outdir if provided in command line
    if kwargs["outdir"]:
        ctx.obj["project"]["outdir"] = kwargs["outdir"]

    # Extract tasks
    ping_tasks = [t for t in ctx.obj["project"]["task_config"] if t["task"] == "ping"]
    profile_tasks = [
        t for t in ctx.obj["project"]["task_config"] if t["task"] == "profile"
    ]
    compare_row_tasks = [
        t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare-row"
    ]
    compare_column_tasks = [
        t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare-column"
    ]
    compare_tasks = [
        t for t in ctx.obj["project"]["task_config"] if t["task"] == "compare"
    ]
    scan_tasks = [t for t in ctx.obj["project"]["task_config"] if t["task"] == "scan"]

    # PingTask
    for tconf in ping_tasks:
        log.debug(f"Executing ping with task profile: {tconf}")
        PingTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            datasources=tconf["datasources"],
        ).execute()

    # ProfileTask
    for tconf in profile_tasks:
        log.debug(f"Executing profile with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "profile_metadata.xlsx")

        try:
            ProfileTask(
                profile=ctx.obj["profile"],
                project=ctx.obj["project"],
                runtime=ctx.obj["runtime"],
                datasources=tconf["datasources"],
                outfile_fqn=outfile_fqn,
                compare=tconf["compare"] if "compare" in tconf else None,
            ).execute()
        except Exception:
            log.error(f"Profiling failed with error: {traceback.format_exc()}")

    # CompareRowTask
    for tconf in compare_row_tasks:
        log.debug(f"Executing compare-row with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "row_comparison.xlsx")

        try:
            CompareRowTask(
                profile=ctx.obj["profile"],
                project=ctx.obj["project"],
                runtime=ctx.obj["runtime"],
                datasources=tconf["datasources"],
                outfile_fqn=outfile_fqn,
                sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
            ).execute()
        except Exception:
            log.error(f"Row comparison failed with error: {traceback.format_exc()}")

    # CompareColumnTask
    for tconf in compare_column_tasks:
        log.debug(f"Executing compare-column with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "column_comparison.xlsx")

        try:
            CompareColumnTask(
                profile=ctx.obj["profile"],
                project=ctx.obj["project"],
                runtime=ctx.obj["runtime"],
                datasources=tconf["datasources"],
                outfile_fqn=outfile_fqn,
                composite=tconf["composite"] if "composite" in tconf else None,
            ).execute()
        except Exception:
            log.error(f"Column comparison failed with errorr: {traceback.format_exc()}")

    # CompareTask
    for tconf in compare_tasks:
        log.debug(f"Executing compare with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )
        outfile_fqn = Path(final_outdir, "comparison.xlsx")

        CompareTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            outfile_fqn=outfile_fqn,
            sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
            composite=tconf["composite"] if "composite" in tconf else None,
        ).execute()

    # ScanTask
    for tconf in scan_tasks:
        log.debug(f"Executing scan with task profile: {tconf}")
        final_outdir = get_final_outdir(
            basedir=ctx.obj["project"]["outdir"],
            task_conf=tconf,
        )

        ScanTask(
            profile=ctx.obj["profile"],
            project=ctx.obj["project"],
            runtime=ctx.obj["runtime"],
            datasources=tconf["datasources"],
            final_outdir=final_outdir,
            compare=tconf["compare"] if "compare" in tconf else None,
            sample_count=tconf["sample_count"] if "sample_count" in tconf else None,
            composite=tconf["composite"] if "composite" in tconf else None,
        ).execute()


if __name__ == "__main__":
    cli()



import logging
import os
import time
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Union

from tulona.config.runtime import RunConfig
from tulona.exceptions import TulonaMissingPropertyError, TulonaUnSupportedTaskError
from tulona.task.base import BaseTask
from tulona.task.compare import CompareTask
from tulona.task.helper import perform_comparison
from tulona.util.excel import dataframes_into_excel
from tulona.util.filesystem import create_dir_if_not_exist
from tulona.util.profiles import extract_profile_name, get_connection_profile
from tulona.util.sql import get_query_output_as_df

log = logging.getLogger(__name__)

DEFAULT_VALUES = {
    "compare_scans": False,
    "sample_count": 20,
    "compare_column_composite": False,
}
META_EXCLUSION = {
    "schemas": ["INFORMATION_SCHEMA", "PERFORMANCE_SCHEMA"],
}


@dataclass
class ScanTask(BaseTask):
    profile: Dict
    project: Dict
    runtime: RunConfig
    datasources: List[str]
    final_outdir: Union[Path, str]
    compare: bool = DEFAULT_VALUES["compare_scans"]
    sample_count: int = DEFAULT_VALUES["sample_count"]
    composite: bool = DEFAULT_VALUES["compare_column_composite"]

    def execute(self):
        log.info(f"Starting task: scan{' --compare' if self.compare else ''}")
        log.debug(f"Full output directory: {self.final_outdir}")
        start_time = time.time()

        scan_result = {}
        ds_name_compressed_list = []
        connection_profile_names = []
        primary_keys = []
        ds_config_list = []
        for ds_name in self.datasources:
            log.info(f"Processing datasource {ds_name}")
            ds_compressed = ds_name.replace("_", "")
            ds_name_compressed_list.append(ds_compressed)
            ds_config = self.project["datasources"][ds_name]
            if "table" in ds_config or "query" in ds_config:
                raise TulonaUnSupportedTaskError(
                    "Scan doesn't work on table/query."
                    " Please use one of the following tasks:"
                    " profile, compare-row, compare-column, compare"
                )
            ds_config_list.append(ds_config)
            scan_result[ds_name] = {}
            scan_result[ds_name]["database"] = {}
            scan_result[ds_name]["schema"] = {}
            if "primary_key" in ds_config:
                primary_keys.append(ds_config["primary_key"])

            connection_profile_name = extract_profile_name(self.project, ds_name)
            connection_profile_names.append(connection_profile_name)
            dbtype = self.profile["profiles"][connection_profile_name]["type"]
            log.debug(
                f"Connection profile: {connection_profile_name} | Database type: {dbtype}"
            )
            scan_result[ds_name]["dbtype"] = dbtype.lower()

            connection_profile = get_connection_profile(self.profile, ds_config)
            conman = self.get_connection_manager(conn_profile=connection_profile)

            # MySQL doesn't have logical database
            if "database" in ds_config:
                database = ds_config["database"]
            elif dbtype.lower() == "mysql":
                database = "def"
            else:
                raise TulonaMissingPropertyError(
                    "Datasource config is missing 'database'."
                )

            # Create output directory
            _ = create_dir_if_not_exist(self.final_outdir)

            # Database scan
            log.debug(f"Performing database scan for: {database}")
            if "schema" in ds_config:
                schemata_query = f"""
                select
                    *
                from
                    {"" if dbtype == "mysql" else database + "."}information_schema.schemata
                where
                    upper(catalog_name) = '{database.upper()}'
                    and upper(schema_name) = '{ds_config["schema"].upper()}'
                """
            else:
                schemata_query = f"""
                select
                    *
                from
                    {"" if dbtype == "mysql" else database + "."}information_schema.schemata
                where
                    upper(catalog_name) = '{database.upper()}'
                    and upper(schema_name) not in (
                        '{"', '".join(META_EXCLUSION['schemas'])}'
                    )
                """
            log.debug(f"Executing query: {schemata_query}")
            dbextract_df = get_query_output_as_df(
                connection_manager=conman, query_text=schemata_query
            )
            log.debug(f"Number of schemas found: {dbextract_df.shape[0]}")

            dbextract_df = dbextract_df.rename(
                columns={c: c.lower() for c in dbextract_df.columns}
            )

            if not self.compare:
                # Writing database scan result
                dbscan_outfile_fqn = Path(
                    self.final_outdir, f"{database.replace('_', '')}_scan.xlsx"
                )
                log.debug(f"Writing db scan result into: {dbscan_outfile_fqn}")
                dataframes_into_excel(
                    sheet_df_map={database: dbextract_df},
                    outfile_fqn=dbscan_outfile_fqn,
                    mode="a" if os.path.exists(dbscan_outfile_fqn) else "w",
                )

            scan_result[ds_name]["database"][database.lower()] = dbextract_df

            # Schema scan
            schema_list = dbextract_df["schema_name"].tolist()
            for schema in schema_list:
                log.debug(f"Performing schema scan for: {database}.{schema}")
                tables_query = f"""
                select
                    *
                from
                    {"" if dbtype == "mysql" else database + "."}information_schema.tables
                where
                    upper(table_catalog) = '{database.upper()}'
                    and upper(table_schema) = '{schema.upper()}'
                """
                log.debug(f"Executing query: {tables_query}")
                schemaextract_df = get_query_output_as_df(
                    connection_manager=conman, query_text=tables_query
                )
                log.debug(f"Number of tables found: {schemaextract_df.shape[0]}")
                schemaextract_df = schemaextract_df.rename(
                    columns={c: c.lower() for c in schemaextract_df.columns}
                )

                if not self.compare:
                    # Writing schema scan result
                    schemascan_outfile_fqn = Path(
                        self.final_outdir,
                        f"{database.replace('_', '')}_{schema.replace('_', '')}_scan.xlsx",
                    )
                    log.debug(
                        f"Writing schema scan result into: {schemascan_outfile_fqn}"
                    )
                    dataframes_into_excel(
                        sheet_df_map={schema: schemaextract_df},
                        outfile_fqn=schemascan_outfile_fqn,
                        mode="a" if os.path.exists(schemascan_outfile_fqn) else "w",
                    )
                scan_result[ds_name]["schema"][
                    f"{database.lower()}.{schema.lower()}"
                ] = schemaextract_df

        if self.compare:
            log.debug("Preparing metadata comparison")

            # Handle primary keys for table comparison
            table_primary_key = list(set(primary_keys))
            if not table_primary_key or len(primary_keys) != len(ds_name_compressed_list):
                log.warning(
                    "Primary key is not specified for one/all"
                    f" comparison candidate datasources: {self.datasources}"
                    " Tulona will attempt to extract it from table metadata."
                )
                table_primary_key = None
            if table_primary_key and len(table_primary_key) > 1:
                log.warning(
                    "Primary key[s] must be same for all candidate datasources for comparison"
                    " otherwise table comparison won't work."
                )
                table_primary_key = None

            # Compare database extracts
            databases = [list(scan_result[k]["database"].keys())[0] for k in scan_result]
            db_frames = [
                list(scan_result[k]["database"].values())[0] for k in scan_result
            ]
            dbtypes = [scan_result[k]["dbtype"] for k in scan_result]
            log.debug(f"Comparing databases: {' vs '.join(databases)}")
            # TODO: schema name, if mentioned in the config, can be different
            # for two data sources but still can be comparison candidate
            # or maybe there is no need
            # outer join will keep both rows
            db_comp = perform_comparison(
                ds_compressed_names=databases,
                dataframes=db_frames,
                on="schema_name",
                how="outer",
                suffixes=ds_name_compressed_list,
                indicator="presence",
            )
            db_comp["presence"] = db_comp["presence"].map(
                {
                    "both": "both",
                    "left_only": ds_name_compressed_list[0],
                    "right_only": ds_name_compressed_list[1],
                }
            )

            # Writing database comparison result
            dbs_compressed = [db.replace("_", "") for db in databases]
            dbcomp_outfile_fqn = Path(
                self.final_outdir, f"{'_'.join(dbs_compressed)}_comparison.xlsx"
            )
            log.debug(f"Writing db scan comparison result into: {dbcomp_outfile_fqn}")
            dataframes_into_excel(
                sheet_df_map={f"db_{'|'.join(databases)}": db_comp},
                outfile_fqn=dbcomp_outfile_fqn,
                mode="a" if os.path.exists(dbcomp_outfile_fqn) else "w",
            )

            # Compare schema extracts: list[list[Dict, Dict]]
            # [
            #     [
            #         {"datasource": "ds1", "database": "db1", "schema": "sc1"},
            #         {"datasource": "ds2", "database": "db2", "schema": "sc2"},
            #     ],
            # ]
            schema_combinations = []
            schemas_from_config = [c["schema"] for c in ds_config_list if "schema" in c]
            if len(schemas_from_config) == 0:
                common_schemas = db_comp[db_comp["presence"] == "both"][
                    "schema_name"
                ].tolist()
                for sc in common_schemas:
                    schema_combinations.append(
                        [
                            {"datasource": ds, "database": db, "schema": sc}
                            for ds, db, sc in zip(
                                ds_name_compressed_list,
                                databases,
                                [sc] * len(ds_name_compressed_list),
                            )
                        ]
                    )
            elif len(schemas_from_config) == len(ds_config_list):
                schema_combinations = [
                    [
                        {"datasource": ds, "database": db, "schema": sc}
                        for ds, db, sc in zip(
                            ds_name_compressed_list, databases, schemas_from_config
                        )
                    ]
                ]
            else:
                log.warning(
                    "Cannot perform schema comparison if schema is specified for some candidate"
                    " datasources but not all. If particular schemas need to be"
                    " scanned and compared, schema must be specified for both datasources."
                    " If schema is not specified for any of them, scan and compare"
                    " will be performed at database level, for all the schemas."
                )

            log.debug(
                f"Number of schema combinations to compare: {len(schema_combinations)}"
            )

            for scombo in schema_combinations:
                log.debug(f"Comparing schema: {scombo}")
                schema_fqns = [f"{cand['database']}.{cand['schema']}" for cand in scombo]
                schema_compressed = [
                    sf.replace(".", "").replace("_", "") for sf in schema_fqns
                ]
                schema_frames = [
                    scan_result[ds_name]["schema"][sf.lower()]
                    for ds_name, sf in zip(self.datasources, schema_fqns)
                ]

                schema_comp = perform_comparison(
                    schema_fqns,
                    schema_frames,
                    on="table_name",
                    how="outer",
                    suffixes=schema_compressed,
                    indicator="presence",
                )
                schema_comp["presence"] = schema_comp["presence"].map(
                    {
                        "both": "both",
                        "left_only": ds_name_compressed_list[0],
                        "right_only": ds_name_compressed_list[1],
                    }
                )

                # Writing schema comparison result
                schemacomp_outfile_fqn = Path(
                    self.final_outdir, f"{'_'.join(schema_compressed)}_comparison.xlsx"
                )
                log.debug(
                    f"Writing schema scan comparison result into: {schemacomp_outfile_fqn}"
                )
                dataframes_into_excel(
                    sheet_df_map={"|".join(schema_compressed): schema_comp},
                    outfile_fqn=schemacomp_outfile_fqn,
                    mode="a" if os.path.exists(schemacomp_outfile_fqn) else "w",
                )

                # Compare tables
                common_tables = schema_comp[schema_comp["presence"] == "both"][
                    "table_name"
                ].tolist()
                log.debug(f"Number of common_tables found: {len(common_tables)}")

                dynamic_project_config = deepcopy(self.project)
                dynamic_project_config["datasources"] = {}
                if "task_config" in dynamic_project_config:
                    dynamic_project_config.pop("task_config")
                for table in common_tables:
                    log.debug(f"Comparing table: {scombo} - {table}")

                    source_map_item = []
                    for cand, typ, cpn in zip(scombo, dbtypes, connection_profile_names):
                        table_ds_config = {
                            "connection_profile": cpn,
                            "schema": cand["schema"],
                            "table": table,
                        }
                        if table_primary_key:
                            table_ds_config["primary_key"] = table_primary_key
                            table_ds_config["compare_column"] = table_primary_key
                        if typ != "mysql":
                            table_ds_config["database"] = cand["database"]

                        sc_comp = cand["schema"].replace("_", "")
                        dyn_ds_name = (
                            f"{cand['datasource']}_{sc_comp}_{table.replace('_', '')}"
                        )
                        dynamic_project_config["datasources"][
                            dyn_ds_name
                        ] = table_ds_config
                        log.debug(
                            f"Datasource: {dyn_ds_name} | Config: {table_ds_config}"
                        )
                        source_map_item.append(dyn_ds_name)

                    table_fqns = [f"{sf}_{table}" for sf in schema_compressed]
                    table_outfile_fqn = Path(
                        self.final_outdir,
                        f"{'_'.join(table_fqns)}_comparison.xlsx",
                    )

                    # Execute CompareTask
                    log.debug(f"Executing CompareTask for: {source_map_item}")
                    CompareTask(
                        profile=self.profile,
                        project=dynamic_project_config,
                        runtime=self.runtime,
                        datasources=source_map_item,
                        outfile_fqn=table_outfile_fqn,
                        sample_count=self.sample_count,
                        composite=self.composite,
                    ).execute()

        exec_time = time.time() - start_time
        compare_flag = " --compare" if self.compare else ""
        log.info(f"Finished task: scan{compare_flag} in {exec_time:.2f} seconds")
